//////////////////////////////////////////////////////////////////////
//#include <stdint.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/i2c-proc.h>
#include <linux/init.h>
#include <video/plmedia.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include "ov7720.h"

#define     LM_DATE     "20070222"
#define     LM_VERSION  "1.0.0.1"
MODULE_LICENSE("GPL");

//////////////////////////////////////////////////////////////////////
// Addresses to scan 
static unsigned short normal_i2c[] = { OV7720_I2C_IOADDR, SENSORS_I2C_END };
static unsigned short normal_i2c_range[] = { SENSORS_I2C_END };
static unsigned int normal_isa[] = { SENSORS_ISA_END };
static unsigned int normal_isa_range[] = { SENSORS_ISA_END };

// Insmod parameters 
SENSORS_INSMOD_1(OV7720);

//////////////////////////////////////////////////////////////////////
static int OV7720AttachAdapter(struct i2c_adapter *adapter);
static int OV7720Detect(struct i2c_adapter *adapter, int address, unsigned short flags, int kind);
static int OV7720DetachClient(struct i2c_client *client);
static void OV7720ProcRead(struct i2c_client *client, int operation, int ctl_name, int *nrels_mag, long *results);
static void OV7720ProcWrite(struct i2c_client *client, int operation, int ctl_name, int *nrels_mag, long *results);
static void OV7720UpdateClient(struct i2c_client *client);
static void OV7720InitClient(struct i2c_client *client);
static int OV7720ReadByte(struct i2c_client *client, int reg);
static int OV7720WriteByte(struct i2c_client *client, int reg, int val);
 
static int OV7720SetFormat(PSENSOR pSensor);
static int OV7720GetFormat(PSENSOR pSensor);
static int OV7720SetSync(PSENSOR pSensor);
static int OV7720GetSync(PSENSOR pSensor);
static void OV7720Restart(PSENSOR pSensor);

static int SensorOpen(struct inode *pinode, struct file *pfile);
static int SeneorRelese(struct inode *pinode, struct file *pfile);
static int SensorIOCTL(struct inode *pinode, struct file *pfile, unsigned int cmd, unsigned long arg);

//////////////////////////////////////////////////////////////////////
static const char OV7720_SHORT_NAME[]="OV7720";
static const char OV7720_FULL_NAME[]="OmniVision OV7720 VGA Sensor";
static const char OV7720_VERSION[]="1.0.0.1";

static struct file_operations SensorFOPS=
{
   owner:      THIS_MODULE,
   open:       SensorOpen,
   release:    SeneorRelese,
   ioctl:      SensorIOCTL,
};
   
//////////////////////////////////////////////////////////////////////
REGISTERINFO OV7720YUVVGADefault[]=
{
    // common setting
    {0x12, 0x80},    // Common Control 7     - Reset all registers to default values
    {0x15, 0x00},    // Common Control 10    - HSYNC, VSYNC = POS
    {0x3d, 0x03},    // Common Control 12    - DC offset compensation for analog process
    {0x65, 0x20},    // DSP Control 2        - Scaling control
    {0x11, 0x01},    // Internal Clock       - 01:30Hz, 00:60Hz

    // vga setting
    {0x17, 0x26},    // Horizontal Frame Start 8 MSBs
    {0x18, 0xa0},    // Horizontal Sensor Size
    {0x19, 0x07},    // Vertical Frame Start 8 MSBs
    {0x1a, 0xf0},    // Vertical Sensor Size
    {0x32, 0x00},    // Image Start and Size - V/H sensor size LSBs, V/H window start LSBs
    {0x29, 0xa0},    // Horizontal Data Output Size 8 MSBs
    {0x2c, 0xf0},    // Vertical Data Output Size 8 MSBs

    // DSP control
    {0x42, 0x7f},    // BLC Blue Channel Target Value
    {0x63, 0xe0},    // AWB Control 0        - AWB gain/calculate enable
    {0x64, 0xff},    // DSP Control 1        - FIFO enable,YUV444 to 422, UV adjust/Color matrix/Interpolation/Gamma/Black&White correction ON
    {0x66, 0x00},    // DSP Control 3
    
    // AGC/AEC/AWB setting
    {0x13, 0xe0},    // Common Control 8     - Disable AGC/AWB/AEC first
    {0x0d, 0x41},    // Common Control 4     - PLL frequency, Full AEC window
    {0x0f, 0xc5},    // Common Control 6     - Auto window setting ON
    {0x14, 0x11},    // Common Control 9     - Auto Gain Ceiling - maximum AGC = 4x
    {0x22, 0x7f},    // Banding Filter Minimum AEC Value
    {0x23, 0x03},    // Banding Filter Maximum Step
    {0x24, 0x40},    // AGC/AEC Stable Region Upper Limit
    {0x25, 0x30},    // AGC/AEC Stable Region Lower Limit
    {0x26, 0xa1},    // AGC/AEC Fast Mode Region
    {0x6b, 0xaa},    // AWB Control 3
    {0x13, 0xef},    // Common Control 8     - Enable AGC/AWB/AEC first

    // Edge Enhancement/Matrix Coefficients
    {0x90, 0x05},    // Edge Enhancement threshold control
    {0x91, 0x01},    // Audo De-noise Threshold Control
    {0x92, 0x03},    // Edge Enhancement Strength low Point Control
    {0x93, 0x00},    // Edge Enhancement Strength high Point Control
    {0x94, 0xb0},    // Matrix Coefficient 1
    {0x95, 0x9d},    // Matrix Coefficient 2
    {0x96, 0x13},    // Matrix Coefficient 3
    {0x97, 0x16},    // Matrix Coefficient 4
    {0x98, 0x7b},    // Matrix Coefficient 5
    {0x99, 0x91},    // Matrix Coefficient 6
    {0x9a, 0x1e},    // Matrix Control       - Matrix double OFF
    {0x9b, 0x08},    // Brightness Control
    {0x9c, 0x20},    // Contrast Control
    {0x9e, 0x81},    // Auto UV Adjust Control 0
    {0xa6, 0x06},    // Special Digital Effect Control
   
    // Gamma setting
    {0x7e, 0x0c},    // Gamma 1
    {0x7f, 0x16},    // Gamma 2
    {0x80, 0x2a},    // Gamma 3
    {0x81, 0x4e},    // Gamma 4
    {0x82, 0x61},    // Gamma 5
    {0x83, 0x6f},    // Gamma 6
    {0x84, 0x7b},    // Gamma 7
    {0x85, 0x86},    // Gamma 8
    {0x86, 0x8e},    // Gamma 9
    {0x87, 0x97},    // Gamma 10
    {0x88, 0xa4},    // Gamma 11
    {0x89, 0xaf},    // Gamma 12
    {0x8a, 0xc5},    // Gamma 13
    {0x8b, 0xd7},    // Gamma 14
    {0x8c, 0xe8},    // Gamma 15
    {0x8d, 0x20},    // Gamma Curve Highest Segment Slope
};

REGISTERINFO OV7720YUVVGA[]=
{
    {0x17, 0x26},    // Horizontal Frame Start 8 MSBs
    {0x18, 0xa0},    // Horizontal Sensor Size
    {0x19, 0x07},    // Vertical Frame Start 8 MSBs
    {0x1a, 0xf0},    // Vertical Sensor Size
    {0x29, 0xa0},    // Horizontal Data Output Size 8 MSBs
    {0x2c, 0xf0},    // Vertical Data Output Size 8 MSBs
};

REGISTERINFO OV7720YUVQVGA[]=
{
    {0x17, 0x3f},    // Horizontal Frame Start 8 MSBs
    {0x18, 0x50},    // Horizontal Sensor Size
    {0x19, 0x03},    // Vertical Frame Start 8 MSBs
    {0x1a, 0x78},    // Vertical Sensor Size
    {0x29, 0x50},    // Horizontal Data Output Size 8 MSBs
    {0x2c, 0x78},    // Vertical Data Output Size 8 MSBs
};

//////////////////////////////////////////////////////////////////////
static ctl_table OV7720_dir_table_template[] = 
{
   {
      OV7720_SYSCTL_READ, 
      "read", 
      NULL, 
      0, 
      0444, 
      NULL, 
      &i2c_proc_real,
      &i2c_sysctl_real, 
      NULL, 
      &OV7720ProcRead
   },
   {
      OV7720_SYSCTL_WRITE, 
      "write", 
      NULL, 
      0, 
      0644, 
      NULL, 
      &i2c_proc_real,
      &i2c_sysctl_real, 
      NULL, 
      &OV7720ProcWrite
   },
   {0}
};

POV7720DATA g_pOV7720Data;
PSENSOR     g_pPLSensor;
static int  OV7720_id = 0;

static struct i2c_driver OV7720Driver =
{
   .owner         = THIS_MODULE,
   .name          = "OV7720 sensor chip driver",   // max. 40 characters limited
   .id            = I2C_DRIVERID_OV7720,           // usually from i2c-id.h
   .flags         = I2C_DF_NOTIFY,                 // always set this value
   .attach_adapter= OV7720AttachAdapter,     
   .detach_client = OV7720DetachClient,
};


//////////////////////////////////////////////////////////////////////
static int OV7720ReadByte(struct i2c_client *client, int reg)
{
   i2c_smbus_write_byte(client, reg);
   return i2c_smbus_read_byte(client);
}

//////////////////////////////////////////////////////////////////////
static int OV7720WriteByte(struct i2c_client *client, int reg, int val)
{
   return i2c_smbus_write_byte_data( client, reg, val&0xFF );
}

//////////////////////////////////////////////////////////////////////
void OV7720ProcRead(struct i2c_client *client, int operation,
   int ctl_name, int *nrels_mag, long *results)
{
   POV7720DATA data = client->data;
   
   if( operation == SENSORS_PROC_REAL_INFO )
      *nrels_mag = 0;
   else if( operation == SENSORS_PROC_REAL_READ ) 
   {
      OV7720UpdateClient(client);
      results[0] = data->wRead;
      *nrels_mag = 1;
   }  
}

//////////////////////////////////////////////////////////////////////
void OV7720ProcWrite( struct i2c_client *client, int operation, int ctl_name, int *nrels_mag, long *results )
{
   POV7720DATA data = client->data;
   
   if( operation == SENSORS_PROC_REAL_INFO )
      *nrels_mag = 0;
   else if( operation == SENSORS_PROC_REAL_READ ) 
   {
      results[0] = data->wReg;
      results[1] = data->wWrite; 
      *nrels_mag = 2;
   } 
   else if( operation == SENSORS_PROC_REAL_WRITE ) 
   {
      if (*nrels_mag == 2) 
      {
         data->wReg = results[0];
         data->wWrite = results[1];
         OV7720WriteByte( client, data->wReg, data->wWrite );
      }
   }
}

//////////////////////////////////////////////////////////////////////
static void HWResetSensor(struct i2c_client *client)
{
   int i;
   
   OV7720WriteByte(client, OV7720REG_COM7, 0x80);  // change to page 0
   OV7720WriteByte(client, OV7720REG_COM7, 0x00);  // set reset bit [0]<--1
   for( i=0; i< sizeof(OV7720YUVVGADefault)/sizeof(OV7720YUVVGADefault[0]); i++ )
      OV7720WriteByte( client, OV7720YUVVGADefault[i].uRegister, OV7720YUVVGADefault[i].uValue );
}

//////////////////////////////////////////////////////////////////////
static int OV7720AttachAdapter(struct i2c_adapter *adapter)
{
   int rc;
   
   printk( "OV7720: AttachAdapter()\n" );
   rc = i2c_detect( adapter, &addr_data, OV7720Detect );
   if( rc < 0 )
      printk("OV7720: Call i2c_detect() failed\n");
   return rc;
}

//////////////////////////////////////////////////////////////////////
int OV7720Detect(struct i2c_adapter *adapter, int address, unsigned short flags, int kind)
{
   struct i2c_client *new_client;
   POV7720DATA data;
   int err = 0, i;
   const char *type_name, *client_name;

   printk( "OV7720: Detect()\n" );

   if( !i2c_check_functionality(adapter, I2C_FUNC_SMBUS_BYTE | I2C_FUNC_SMBUS_BYTE_DATA) )
      goto ERROR0;

   if( !(new_client = kmalloc(sizeof(struct i2c_client)+sizeof(OV7720DATA)+sizeof(SENSOR), GFP_KERNEL)) ) 
   {
      err = -ENOMEM;
      goto ERROR0;
   }

   g_pOV7720Data= data = (OV7720DATA *) (new_client + 1);
   g_pPLSensor = data->pSensor =(PSENSOR)(data+1);
   new_client->addr = address;
   new_client->data = data;
   new_client->adapter = adapter;
   new_client->driver = &OV7720Driver;
   new_client->flags = 0;

   if( kind <= 0 ) 
   {
      if( address == OV7720_I2C_IOADDR && 
          OV7720_PID==OV7720ReadByte(new_client, OV7720REG_PID) &&
          OV7720_VER==OV7720ReadByte(new_client, OV7720REG_VER) )
            kind = OV7720;
      else 
      {
         printk("OV7720: Unknown sensor\n");
         goto ERROR1;
      }
   }
   else if( kind == OV7720 ) 
   {
      type_name = "OV7720";
      client_name = "OV7720 chip";
   } 
   else 
   {
      goto ERROR1;
   }

   // Fill in the remaining client fields and put it into the global list 
   strcpy( new_client->name, client_name );

   new_client->id = OV7720_id++;
   data->cValid = 0;
   init_MUTEX( &data->UpdateLock );

   // Tell the I2C layer a new client has arrived
   if( (err = i2c_attach_client(new_client)) )
      goto ERROR3;

   // Register a new directory entry with module sensors
   if( (i = i2c_register_entry(new_client, type_name, OV7720_dir_table_template)) < 0 ) 
   {
      err = i;
      goto ERROR4;
   }
   data->nSysCtlID = i;

   //Initialize the OV7720D chip
   OV7720InitClient(new_client);
   
   return 0;

// OK, this is not exactly good programming practice, usually. But it is
// very code-efficient in this case.

ERROR4:
   i2c_detach_client(new_client);
ERROR3:
ERROR1:
   kfree(new_client);
ERROR0:
   return err;
}


//////////////////////////////////////////////////////////////////////
static int OV7720DetachClient(struct i2c_client *client)
{
   int err;

   printk( "OV7720: DetachClient()\n" );
   
   i2c_deregister_entry(((POV7720DATA) (client->data))-> nSysCtlID);

   if( (err = i2c_detach_client(client)) ) 
   {
      printk("OV7720: Deregistration failed!\n");
      return err;
   }
   kfree(client);
   return 0;
}

//////////////////////////////////////////////////////////////////////
static void OV7720InitClient(struct i2c_client *client)
{
   POV7720DATA    data = client->data;
   int            i;

   printk( "OV7720: InitClient()\n" );
   
   data->pSensor->nType       = TYPE_SENSOR; 
   data->pSensor->pI2CClient  = client;
   data->pSensor->szShortName = OV7720_SHORT_NAME;
   data->pSensor->szFullName  = OV7720_FULL_NAME;
   data->pSensor->szVersion   = OV7720_VERSION;
   data->pSensor->PrivateData = data;
   data->pSensor->Restart     = OV7720Restart;
   data->pSensor->SetFormat   = OV7720SetFormat;
   data->pSensor->GetFormat   = OV7720GetFormat;
   data->pSensor->SetSync     = OV7720SetSync;
   data->pSensor->GetSync     = OV7720GetSync;
   
   data->pSensor->SyncInfo.nHSyncStart    = 0;
   data->pSensor->SyncInfo.nVSyncStart    = 0;
   data->pSensor->SyncInfo.byITUBT656     = 0; 
   data->pSensor->SyncInfo.byHRefSelect   = 1;
   data->pSensor->SyncInfo.nHSyncPhase    = 1;
   data->pSensor->SyncInfo.byHSyncState   = 0;
   data->pSensor->SyncInfo.byVSyncState   = 0;
   data->pSensor->SyncInfo.byValidEnable  = 0;
   data->pSensor->SyncInfo.byValidState   = 0;
   
   data->pSensor->Format.nIsVarioPixel    = 0;
   data->pSensor->Format.nInterlaced      = 0;

   data->DevFS = devfs_register (NULL, "sensor", DEVFS_FL_AUTO_DEVNUM, 0, 0,   // auto major and minor number
                                 S_IFCHR | S_IWUSR | S_IRUSR, &SensorFOPS, NULL);
      
   HWResetSensor(client);
   
   for( i=0; i< sizeof(OV7720YUVVGADefault)/sizeof(OV7720YUVVGADefault[0]); i++ )
      OV7720WriteByte( client, OV7720YUVVGADefault[i].uRegister, OV7720YUVVGADefault[i].uValue );

   printk( "OV7720: CMOS sensor %s is avaliable now\n", data->pSensor->szShortName );
}

//////////////////////////////////////////////////////////////////////
static void OV7720UpdateClient(struct i2c_client *client)
{
   OV7720DATA *data = client->data;

   down(&data->UpdateLock);
   if( (jiffies - data->uLastUpdated > 5*HZ) || (jiffies < data->uLastUpdated) || !data->cValid ) 
   {
      data->wRead = i2c_smbus_read_byte(client); 
      data->uLastUpdated = jiffies;
      data->cValid = 1;
   }
   up(&data->UpdateLock);
}

//////////////////////////////////////////////////////////////////////
static int __init sm_OV7720_init(void)
{
	printk("OV7720: init(): version %s (%s)\n", LM_VERSION, LM_DATE);
   return i2c_add_driver(&OV7720Driver);
}

//////////////////////////////////////////////////////////////////////
static void __exit sm_OV7720_exit(void)
{
   i2c_del_driver(&OV7720Driver);
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
static int SensorOpen(struct inode *pinode, struct file *pfile)
{
   return 0;
}

//////////////////////////////////////////////////////////////////////
static int SeneorRelese(struct inode *pinode, struct file *pfile)
{
   return 0;
}

//////////////////////////////////////////////////////////////////////
static int SensorIOCTL(struct inode *pinode, struct file *pfile, unsigned int cmd, unsigned long arg)
{

   struct i2c_client *pClient = g_pPLSensor->pI2CClient;
   REGISTERINFO      reg;
   
   switch (cmd)
   {
      case PLSENSOR_WRITE_REGISTER:
         copy_from_user( (void*)&reg , (void*)arg, sizeof(REGISTERINFO) );
         OV7720WriteByte( pClient, reg.uRegister, reg.uValue );
         break;
      case PLSENSOR_READ_REGISTER:
         copy_from_user( (void*)&reg , (void*)arg, sizeof(REGISTERINFO) );
         reg.uValue = OV7720ReadByte( pClient, reg.uRegister );
         copy_to_user( (void*)arg, (void*)&reg, sizeof(REGISTERINFO) );
         break;
      case PLSENSOR_WRRD_REGISTER:
         copy_from_user( (void*)&reg , (void*)arg, sizeof(REGISTERINFO) );
         OV7720WriteByte( pClient, reg.uRegister, reg.uValue );
         reg.uValue = OV7720ReadByte( pClient, reg.uRegister );
         copy_to_user( (void*)arg, (void*)&reg, sizeof(REGISTERINFO) );
         break;
   }
   return 0;
}

//////////////////////////////////////////////////////////////////////
static int OV7720SetFormat(PSENSOR pSensor)
{
   int               bIsNotSXVGA=1;
   struct i2c_client *pClient=pSensor->pI2CClient;
   int i;
       
   printk( "OV7720: SetFormat(): %d x %d\n", pSensor->Format.nCropWidth, pSensor->Format.nCropHeight );

   if( pSensor->Format.nCropWidth == 320 && pSensor->Format.nCropHeight == 240 )
   { // QVGA
      OV7720WriteByte( pClient, OV7720REG_COM7, 0x40 );
      for( i=0; i< sizeof(OV7720YUVQVGA)/sizeof(OV7720YUVQVGA[0]); i++ )
         OV7720WriteByte( pClient, OV7720YUVQVGA[i].uRegister, OV7720YUVQVGA[i].uValue );
       
   }
   else // unknown width and height...!!
   { // default set as VGA 
      OV7720WriteByte( pClient, OV7720REG_COM7, 0x00 );
      for( i=0; i< sizeof(OV7720YUVVGA)/sizeof(OV7720YUVVGA[0]); i++ )
         OV7720WriteByte( pClient, OV7720YUVVGA[i].uRegister, OV7720YUVVGA[i].uValue );
   }
   
   pSensor->Format.uHSTART = 0;
   pSensor->Format.uVSTART = 0;
   pSensor->Format.uHSTOP = pSensor->Format.uHSTART + (pSensor->Format.nCropWidth << bIsNotSXVGA);
   pSensor->Format.uVSTOP  =  pSensor->Format.uVSTART+pSensor->Format.nCropHeight;
   pSensor->Format.nInterlaced = 0;
   pSensor->Format.nFPS = 30;    
   
   return 0;
}

//////////////////////////////////////////////////////////////////////
static int OV7720GetFormat(PSENSOR pSensor)
{
   printk( "OV7720: GetFormat()\n" );
   return 0;   
}

//////////////////////////////////////////////////////////////////////
static int OV7720SetSync(PSENSOR pSensor)
{
   printk( "OV7720: SetSync()\n" );
   pSensor->SyncInfo.byITUBT656 = 0;
   pSensor->SyncInfo.byHRefSelect = 1;
	pSensor->SyncInfo.nHSyncPhase = 1;
   return 0;
}

//////////////////////////////////////////////////////////////////////
static int OV7720GetSync(PSENSOR pSensor)
{
   printk( "OV7720: GetSync()\n" );
   return 0;   
}

//////////////////////////////////////////////////////////////////////
static void OV7720Restart(PSENSOR pSensor)
{
   printk( "OV7720: Restart()\n" );
   HWResetSensor(pSensor->pI2CClient);
}

//////////////////////////////////////////////////////////////////////
MODULE_AUTHOR("Jon.Lee");
MODULE_DESCRIPTION("OV7720 driver");
EXPORT_SYMBOL(g_pPLSensor);
module_init(sm_OV7720_init);
module_exit(sm_OV7720_exit);

//////////////////////////////////////////////////////////////////////

